# coding: utf-8
"""
TODO: add a docstring.
"""

import os.path
from copy import copy

from state import State
from context import Context
from rendering import render
from loading import load_file, load_template
from utils import make_unicode, html_escape

__version__ = '0.1.4'
__all__ = ['Context',
           'State',
           'html_escape',
           'load_file',
           'load_template',
           'make_unicode',
           'render',
           'template']

template_globals = {}

def template(relative_path, *args, **kwargs):
    """ A decorator for easily rendering templates.
    Use as follows:

    main.py:
        from mustache import (template)

        @template('../tests/static/say_hello.html')
        def index():
            context = {'name' : 'world'}
            partials = {}
            return context, partials

        if __name__=="__main__":
            print index()

    static/say_hello.html:

        <h1> Hello, {{name}}! </h1>

    from the command line:

        > python main.py
        <h1> Hello, world! </h1>
    """
    directory, filename = os.path.split(relative_path)

    partials_dir = os.path.abspath(directory)
    name, ext = os.path.splitext(filename)

    state = State(partials_dir=directory, extension=ext, *args, **kwargs)
    template = load_template(name, directory, ext, state.encoding, state.encoding_error)

    def wrapper(fn):
        def render_template(*args, **kwargs):
            res = fn(*args, **kwargs)
            if isinstance(res, tuple):
                if len(res) == 2:
                    (new_context, partials) = res
                elif len(res) == 1:
                    (new_context, partials) = (res[0], {})
            elif isinstance(res, dict):
                (new_context, partials) = (res, {})
            else:
                (new_context, partials) = ({}, {})
            context = copy(template_globals)
            context.update(new_context)
            return render(template, context, partials, state)
        return render_template
    return wrapper




# coding: utf-8
import re
from os.path import split

from context import Context
from loading import load_template
from state import State
from utils import make_unicode, html_escape

def get_match_info(template, match, state):
    """
    Given a template and a regex match within said template, return a
    dictionary of information about the match to be used to help parse the
    template.
    """
    info = match.groupdict()

    # Put special delimiter cases in terms of normal ones
    if info['change']:
        info.update({
            'tag_type' : '=',
            'tag_key' : info['delims'],
        })
    elif info['raw']:
        info.update({
            'tag_type' : '&',
            'tag_key' : info['raw_key'],
        })

    # Rename the important match variables for convenience
    tag_start = match.start()
    tag_end = match.end()
    tag_type = info['tag_type']
    tag_key = info['tag_key']
    lead_wsp = info['lead_wsp']
    end_wsp = info['end_wsp']

    begins_line = (tag_start == 0) or (template[tag_start-1] in state.eol_chars)
    ends_line = (tag_end == len(template) or
                 template[tag_end] in state.eol_chars)
    interpolating = (tag_type in ('', '&'))
    standalone = (not interpolating) and begins_line and ends_line

    if end_wsp:
        tag_end -= len(end_wsp)
    if standalone:
        template_length = len(template)
        # Standalone tags strip exactly one occurence of '\r', '\n', or '\r\n'
        # from the end of the line.
        if tag_end < len(template) and template[tag_end] == '\r':
            tag_end += 1
        if tag_end < len(template) and template[tag_end] == '\n':
            tag_end += 1
    elif lead_wsp:
        tag_start += len(lead_wsp)
        lead_wsp = ''

    info.update({
        'tag_start' : tag_start,
        'tag_end' : tag_end,
        'tag_type' : tag_type,
        'tag_key' : tag_key,
        'lead_wsp' : lead_wsp,
        'end_wsp' : end_wsp,
        'begins_line' : begins_line,
        'ends_line' : ends_line,
        'interpolating' : interpolating,
        'standalone' : standalone,
    })
    return info

def get_tag_context(name, state):
    """
    Given a tag name, return its associated value as defined in the current
    context stack.
    """
    new_contexts = 0
    ctm = None
    while True:
        try:
            ctx_key, name = name.split('.', 1)
            ctm = state.context.get(ctx_key)
        except ValueError:
            break
        if not ctm:
            break
        else:
            state.context.push(ctm)
            new_contexts += 1

    ctm = state.context.get(name)

    return new_contexts, ctm

def section_end_info(template, tag_key, state, index):
    """
    Given the tag key of an opening section tag, find the corresponding closing
    tag (if it exists) and return information about that match.
    """

    state.section.push(tag_key)
    match = None
    matchinfo = None
    search_index = index

    while state.section:
        match = state.tag_re.search(template, search_index)
        if not match:
            raise Exception("Open section %s never closed" % tag_key)

        matchinfo = get_match_info(template, match, state)

        # If we find a new section tag, add it to the stack and keep going
        if matchinfo['tag_type'] in ('#', '^'):
            state.section.push(matchinfo['tag_key'])
        # If we find a closing tag for the current section, 'close' it by
        # popping the stack
        elif matchinfo['tag_type'] == '/':
            if matchinfo['tag_key'] == state.section():
                state.section.pop()
            else:
                raise Exception(
                    'Unexpected section end: received %s, expected {{/%s}}' % (
                        repr(match.group(0)), tag_key))
        search_index = matchinfo['tag_end']

    return matchinfo


def render(template, context, partials={}, state=None):
    """ Renders a given mustache template, with sane defaults. """
    # Create a new state by default
    state = state or State()

    # Add context to the state dict
    if isinstance(context, Context):
        state.context = context
    else:
        state.context = Context(context)

    # Add any partials to the state dict
    if partials:
        state.partials.push(partials)

    # Render the rendered template
    return __render(make_unicode(template), state)

def __render(template, state, index=0):
    """
    Given a /template/ string, a parser /state/, and a starting
    offset (/index/), return the rendered version of the template.
    """
    # Find a Match
    match = state.tag_re.search(template, index)
    if not match:
        return template[index:]

    info = get_match_info(template, match, state)
    _pre = template[index : info['tag_start']] # template before the tag
    _tag = template[info['tag_start'] : info['tag_end']] # tag
    _continue = info['tag_end'] # the index at which to continue

    # Comment
    if info['tag_type'] == '!':
        # Comments are removed from output
        repl = ""

    # Delimiter change
    elif info['tag_type'] == '=':
        # Delimiters are changed; the tag is rendered as ""
        delimiters = re.split(r'\s*', info['tag_key'])
        new_tags = state.tags(_copy=True)
        new_tags['otag'], new_tags['ctag'] = map(re.escape, delimiters)
        state.push_tags(new_tags)
        repl = ""

    # Plain tag
    elif info['tag_type'] == '':
        repl = __render_tag(info, state)

    # Raw tag (should not be escaped)
    elif info['tag_type'] == '&':
        state.escape.push(False)
        repl = __render_tag(info, state)
        state.escape.pop()

    # Partial
    elif info['tag_type'] == '>':
        partial_name = info['tag_key']
        partial_template = None
        new_dir = None
        lead_wsp = re.compile(r'^(.)', re.M)
        repl = ''
        try:
            # Cached
            partial_template = state.partials()[partial_name]
        except (KeyError, IndexError):
            try:
                # Load the partial template from a file (if it exists)
                new_dir, filename = split(partial_name)
                if new_dir:
                    state.partials_dir.push(new_dir)

                partial_template = load_template(filename, state.abs_partials_dir, state.extension,
                                                 state.encoding, state.encoding_error)
            except (IOError):
                pass

        if partial_template:
            # Preserve indentation
            if info['standalone']:
                partial_template = lead_wsp.sub(info['lead_wsp']+r'\1', partial_template)

            # Update state
            state.partials.push(state.partials()) # XXX wtf is this shit?
            state.push_tags(state.default_tags)

            # Render the partial
            repl = __render(partial_template, state)

            # Restore state
            state.partials.pop()
            state.pop_tags()
            if new_dir:
                state.partials_dir.pop()

    # Section

    # TODO(peter): add a stop= index to __render so that template_to_inner does
    # not need to be constructed with [:] indexing, which is extremely
    # expensive.
    elif info['tag_type'] in ('#', '^'):
        otag_info = info
        ctag_info = section_end_info(template, info['tag_key'], state, _continue)

        # Don't want to parse beyond the end of the inner section, but
        # must include information on prior contents so that whitespace
        # is preserved correctly and inner tags are not marked as standalone.
        inner_start = otag_info['tag_end']
        inner_end = ctag_info['tag_start']
        _continue = ctag_info['tag_end']

        template_with_inner = template[:inner_end]

        new_contexts, ctm = get_tag_context(otag_info['tag_key'], state)
        truthy = otag_info['tag_type'] == '#'

        #if ctm is not None:
        if ctm:
            # If there's a match and it's callable, feed it the inner template
            if callable(ctm):
                template_to_inner = template[:inner_start]
                inner = template[inner_start:inner_end]
                template_with_inner = template_to_inner + make_unicode(ctm(inner))

            # Make the context list an iterable from the ctm
            if not hasattr(ctm, '__iter__') or isinstance(ctm, dict):
                ctx_list = [ctm]
            else:
                ctx_list = ctm
        # If there's no match, there are no new contexts
        else:
            ctx_list = [False]

        # If there are new contexts and the section is truthy, or if
        # there are no new contexts and the section is falsy, render
        # the contents
        repl_stack = []
        for ctx in ctx_list:
            if (truthy and ctx) or (not truthy and not ctx):
                state.context.push(ctx)
                repl_stack.append(
                    __render(template_with_inner, state, inner_start))

            else:
                break

        repl = ''.join(repl_stack)
        for i in xrange(new_contexts): state.context.pop()
    else:
        raise Exception("found unpaired end of section tag!")

    return u''.join((
        _pre, make_unicode(repl), __render(template, state, _continue)))

def __render_tag(info, state):
    """ Render an individual tag by making the appropriate replacement within
    the current context (if any). """
    new_contexts, context_match = get_tag_context(info['tag_key'], state)
    replacement = ''

    if context_match or context_match == 0:
        replacement = context_match
    elif info['tag_key'] == '.':
        replacement = state.context()
    else:
        replacement = ''

    # Call all callables / methods / lambdas / functions
    if replacement and callable(replacement):
        replacement = make_unicode(replacement())

        state.push_tags(state.default_tags)
        replacement = __render(template=replacement, state=state)
        state.pop_tags()

    for i in xrange(new_contexts): state.context.pop()

    if state.escape():
        return html_escape(replacement)
    return replacement




# coding: utf-8
from copy import copy

class Nothing(object):
    """ A sentinal object for when things don't exist. """
    pass

class Stack(list):
    """ A simple wrapper for lists, providing nicer semantics for accessing the
    last element and alternate syntax for appending items. """
    def __init__(self, obj=Nothing, *args, **kwargs):
        super(list, self).__init__(*args, **kwargs)
        if obj is not Nothing:
            self.append(obj)

    def __call__(self, _copy=False):
        if not _copy:
            return self[-1]
        return copy(self[-1])

    def push(self, *args, **kwargs):
        return self.append(*args, **kwargs)




# coding: utf-8
import re
from os.path import join, abspath

from copy import copy
from stack import Stack
from context import Context

class State(object):
    """
    An object to keep track of the parser's state as it works its way
    through a string to be parsed. Uses sane defaults (that comply with the
    mustache spec), but allows for customization.
    """

    def __init__(self,
                 partials_dir='.',
                 extension='mustache',
                 encoding='utf-8',
                 encoding_error='xmlcharrefreplace'):
        # Partials
        self.partials_dir = Stack(partials_dir)
        self.extension = extension

        # Encoding
        self.encoding = encoding
        self.encoding_error = encoding_error

        # State
        self.partials = Stack({})
        self.section = Stack()
        self.escape = Stack(True)
        self.context = Context()

        # Parsing
        self.tags = Stack({
            'otag' : re.escape('{{'),    # tag opening
            'ctag' : re.escape('}}'),    # tag closing
            'types' : '|'.join(('#', '/', '^', '&', '!', '>')) # tag modifiers
        })
        self.default_tags = copy(self.tags())
        self.eol_chars = ('\r', '\n')
        self.raw_tag_re = r"""
            (?P<lead_wsp>[\ \t]*) # leading whitespace
            %(otag)s \s*
            (?:
                (?P<change>=) \s* (?P<delims>.+?)   \s* = |           # change delimiters OR
                (?P<raw>{)    \s* (?P<raw_key>.+?) \s* } |            # raw format (no escaping) OR
                (?P<tag_type>[%(types)s]?)  \s* (?P<tag_key>[\s\S]+?) # other tag types
            )
            \s* %(ctag)s (?P<end_wsp>[\ \t]*)"""
        self.re_flags = re.MULTILINE|re.VERBOSE|re.DOTALL|re.UNICODE
        self.tag_re = None # replaced by update_tag_re
        self.update_tag_re()


    def compile_tag_re(self, tags):
        """
        Return the regex used to look for Mustache tags compiled to work with
        specific opening tags, close tags, and tag types.
        """
        return re.compile(self.raw_tag_re % tags, self.re_flags)

    def update_tag_re(self):
        """
        Update the compiled tag regex to use the tags on the top of the tag
        stack. This should be called any time the tag stack is changed.
        """
        self.tag_re = self.compile_tag_re(self.tags())

    def push_tags(self, tags):
        """
        Pushes the new tags to the top of the tag state and updates the
        compiled tag regex.
        """
        self.tags.push(tags)
        self.update_tag_re()

    def pop_tags(self):
        """
        Pops the current tags from the top of the tag state and updates the
        compiled tag regex.
        """
        self.tags.pop()
        self.update_tag_re()

    @property
    def abs_partials_dir(self):
        """
        Returns the absolute path to the current partials directory based on
        the current stack of relative partial directories.
        """
        return abspath(join(*self.partials_dir))




# coding: utf-8
from stack import Stack

class Context(object):
    """ A wrapper around a Stack of dictionaries, this object is used to hold
    variable information for use in templates. Every {{key}} in a template is
    looked up within a Context. """

    default = '' # what to return when no match is found (according to spec)
    def __init__(self, context=None, **kwargs):
        context = context or {}
        context.update(kwargs)
        self.stack = Stack(context)

    def push(self, key):
        return self.stack.push(key)

    def pop(self, index=-1):
        return self.stack.pop(index)

    def __call__(self, *args, **kwargs):
        return self.stack(*args, **kwargs)

    def get(self, key, default=''):
        default = default or getattr(self, 'default', default)
        return self.get_or_attr(key, default)

    def get_or_attr(self, key, default):
        for ctx in reversed(self.stack):
            try:
                return ctx[key]
            except (KeyError, TypeError):
                try:
                    return getattr(ctx, key)
                except AttributeError:
                    pass
        return default

    def __getitem__(self, key):
        return self.get(key)

    def __repr__(self):
        return '%s%s' % (self.__class__.__name__, tuple(self.stack))

    def __str__(self):
        return str(self.stack)

    def __unicode__(self):
        return unicode(self.stack)




# coding: utf-8
from os.path import split, splitext, extsep, join, abspath, exists
from utils import make_unicode

DEFAULT_EXTENSION = 'html'
DEFAULT_DIRECTORY = 'static/templates'
DEFAULT_ENCODING = 'utf-8'
DEFAULT_ERRORS = 'xmlcharrefreplace'

def read(path):
    """ Return the contents of a file as a byte string. """
    try:
      f = open(path, 'rb')
      return f.read()
    finally:
      f.close()

def read_unicode(path, encoding, encoding_errors):
    """ Return the contents of a file as a unicode string. """
    try:
      f = open(path, 'rb')
      return make_unicode(f.read(), encoding, encoding_errors)
    finally:
      f.close()

def get_abs_template_path(template_name, directory, extension):
    """ Given a template name, a directory, and an extension, return the
    absolute path to the template. """
    # Get the relative path
    relative_path = join(directory, template_name)
    file_with_ext = template_name

    if extension:
        # If there is a default extension, but no file extension, then add it
        file_name, file_ext = splitext(file_with_ext)
        if not file_ext:
            file_with_ext = extsep.join(
                (file_name, extension.replace(extsep, '')))
            # Rebuild the relative path
            relative_path = join(directory, file_with_ext)

    return abspath(relative_path)

def load_file(path, encoding, encoding_errors):
    """ Given an existing path, attempt to load it as a unicode string. """
    abs_path = abspath(path)
    if exists(abs_path):
        return read_unicode(abs_path, encoding, encoding_errors)
    raise IOError('File %s does not exist' % (abs_path))

def load_template(name, directory, extension, encoding, encoding_errors):
    """ Load a template and return its contents as a unicode string. """
    abs_path = get_abs_template_path(name, directory, extension)
    return load_file(abs_path, encoding, encoding_errors)




from distutils.core import setup
setup(
    name = 'mustache',
    packages = ['mustache'],
    version = '0.1.4',
    description = 'Mustache templating in Python',
    author = 'Peter Downs',
    author_email = 'peterldowns@gmail.com',
    url = 'https://github.com/peterldowns/python-mustache',
    download_url = 'https://github.com/peterldowns/python-mustache/tarball/v0.1.4',
    install_requirements = [],
    extras_require = {
      'test' : ['simplejson'],
    },
    keywords = [
        'templating',
        'template',
        'mustache',
        'web'],
    classifiers = [
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup'],
)



# coding: utf-8
from cgi import escape

def make_unicode(s, encoding='utf-8', encoding_errors='strict'):
    """ Return the unicode version of an input. """
    if not isinstance(s, unicode):
        if not isinstance(s, basestring):
            return unicode(str(s), encoding, encoding_errors)
        return unicode(s, encoding, encoding_errors)
    return s

def html_escape(s, encoding='utf-8', encoding_errors='strict'):
    """ Return the HTML-escaped version of an input. """
    return escape(make_unicode(s, encoding, encoding_errors), quote=True)

